@@ -2,10 +2,17 @@ import pytest
 import itertools
 import grblas
 import numpy as np
+import pickle
 from grblas import Matrix, Vector, Scalar
 from grblas import unary, binary, monoid, semiring
 from grblas import dtypes
-from grblas.exceptions import IndexOutOfBound, DimensionMismatch, OutputNotEmpty, InvalidValue
+from grblas.exceptions import (
+    IndexOutOfBound,
+    DimensionMismatch,
+    OutputNotEmpty,
+    InvalidValue,
+    DomainMismatch,
+)
 
 
 @pytest.fixture
@@ -79,6 +86,9 @@ def test_from_values():
     assert C3.nvals == 2  # duplicates were combined
     assert C3.dtype == int
     assert C3[1, 1].value == 6  # 2*3
+    C3monoid = Matrix.from_values([0, 1, 1], [2, 1, 1], [1, 2, 3], nrows=10, dup_op=monoid.times)
+    assert C3.isequal(C3monoid)
+
     with pytest.raises(ValueError, match="Duplicate indices found"):
         # Duplicate indices requires a dup_op
         Matrix.from_values([0, 1, 1], [2, 1, 1], [True, True, True])
@@ -159,6 +169,12 @@ def test_build(A):
     A.clear()
     with pytest.raises(IndexOutOfBound):
         A.build([0, 11], [0, 0], [1, 1])
+    B = Matrix.new(int, nrows=2, ncols=2)
+    B.build([0, 11], [0, 0], [1, 1], nrows=12)
+    assert B.isequal(Matrix.from_values([0, 11], [0, 0], [1, 1], ncols=2))
+    C = Matrix.new(int, nrows=2, ncols=2)
+    C.build([0, 0], [0, 11], [1, 1], ncols=12)
+    assert C.isequal(Matrix.from_values([0, 0], [0, 11], [1, 1], nrows=2))
 
 
 def test_extract_values(A):
@@ -304,8 +320,8 @@ def test_ewise_mult(A):
     assert C.isequal(result)
     C() << A.ewise_mult(B, monoid.times)
     assert C.isequal(result)
-    C << A.ewise_mult(B, semiring.plus_times)
-    assert C.isequal(result)
+    with pytest.raises(TypeError, match="Expected type: BinaryOp, Monoid"):
+        A.ewise_mult(B, semiring.plus_times)
 
 
 def test_ewise_add(A):
@@ -324,8 +340,10 @@ def test_ewise_add(A):
     assert C.isequal(result)
     C << A.ewise_add(B, monoid.max)
     assert C.isequal(result)
-    C << A.ewise_add(B, semiring.max_minus)
+    C << A.ewise_add(B, binary.max)
     assert C.isequal(result)
+    with pytest.raises(TypeError, match="Expected type: Monoid"):
+        A.ewise_add(B, semiring.max_minus)
 
 
 def test_extract(A):
@@ -980,22 +998,51 @@ def test_apply_binary(A):
     with pytest.raises(TypeError, match="Cannot provide both"):
         A.apply(binary.plus, left=1, right=1)
 
+    # allow monoids
+    w1 = A.apply(binary.plus, left=1).new()
+    w2 = A.apply(monoid.plus, left=1).new()
+    w3 = A.apply(monoid.plus, right=1).new()
+    assert w1.isequal(w2)
+    assert w1.isequal(w3)
+
 
 def test_reduce_row(A):
     result = Vector.from_values([0, 1, 2, 3, 4, 5, 6], [5, 12, 1, 6, 7, 1, 15])
     w = A.reduce_rows(monoid.plus).new()
     assert w.isequal(result)
+    w2 = A.reduce_rows(binary.plus).new()
+    assert w2.isequal(result)
+
+
+def test_reduce_row_udf(A):
+    result = Vector.from_values([0, 1, 2, 3, 4, 5, 6], [5, 12, 1, 6, 7, 1, 15])
+    binop = grblas.operator.BinaryOp.register_anonymous(lambda x, y: x + y)
+    with pytest.raises(DomainMismatch):
+        # Although allowed by the spec, SuiteSparse doesn't like user-defined binarops here
+        A.reduce_rows(binop).new()
+    # If the user creates a monoid from the binop, then we can use the monoid instead
+    monoid = grblas.operator.Monoid.register_anonymous(binop, 0)
+    w = A.reduce_rows(binop).new()
+    assert w.isequal(result)
+    w2 = A.reduce_rows(monoid).new()
+    assert w2.isequal(result)
 
 
 def test_reduce_column(A):
     result = Vector.from_values([0, 1, 2, 3, 4, 5, 6], [3, 2, 9, 10, 11, 8, 4])
     w = A.reduce_columns(monoid.plus).new()
     assert w.isequal(result)
+    w2 = A.reduce_columns(binary.plus).new()
+    assert w2.isequal(result)
 
 
 def test_reduce_scalar(A):
     s = A.reduce_scalar(monoid.plus).new()
     assert s == 47
+    assert A.reduce_scalar(binary.plus).value == 47
+    with pytest.raises(TypeError, match="Expected type: Monoid"):
+        A.reduce_scalar(binary.minus)
+
     # test dtype coercion
     assert A.dtype == dtypes.INT64
     s = A.reduce_scalar().new(dtype=float)
@@ -1633,3 +1680,10 @@ def test_wait(A):
     A2 = A.dup()
     A2.wait()
     assert A2.isequal(A)
+
+
+def test_pickle(A):
+    s = pickle.dumps(A)
+    A2 = pickle.loads(s)
+    assert A.isequal(A2, check_dtype=True)
+    assert A.name == A2.name
